#include <cstdio>
#include <cstdlib>
#include <vector>
#include <cmath>

void usage()
{
	printf(
"Usage: .exe rom fixed_rom [log] [fmscale] [psgscale]\n"
"\n rom       - rom with GEMS driver"
"\n fixed_rom - output path (automatically overwrited)"
"\n log       - any parameter to turn on log"
"\n fmscale   - multiplicator for fm frequences (1.25)"
"\n psgscale  - multiplicator for psg frequences (1.5)"
"\n"
"\n"
"Author: r57shell@uralweb.ru\n"
"Last Update: 14.07.2013\n");
	exit(0);
} 

double ftbl[13]=
{
261.63,
277.18,
293.66,
311.13,
329.63,
349.23,
369.99,
392.00,
415.30,
440.00,
466.16,
493.88,
523.25
};
unsigned short fmftbl[13];
unsigned short psgftbl[64];
std::vector<unsigned char> rom;
bool use_log = false;
#define lprintf if (use_log) printf

void build_fmftbl(double scale)
{
	lprintf("fmftbl: ; scale %lf\n",scale);
	for (int i=0; i<13; ++i)
	{
		// (144 * fnote * 2^20) / 2^3 /FM = F (1 )
		fmftbl[i]=floor(144*ftbl[i]*scale*1024*128/7670000+0.5);
		lprintf(" dw %d\n",(int)fmftbl[i]);
	}
}

double fmscale = 1.25; //2637.0/2093.0;

bool fix_fmftbl()
{
	// build fmftbl like original
	build_fmftbl(1);

	// find fmftbl
	for (int i=0; i+sizeof(fmftbl)-1<rom.size(); ++i)
	{
		int j, n=sizeof(fmftbl)/sizeof(fmftbl[0]);
		for (j=0; j<n; ++j)
			if (rom[i+j*2  ] != (fmftbl[j]&0xFF)
			 || rom[i+j*2+1] != ((fmftbl[j]>>8)&0xFF))
				break;
		if (j == n) // found
		{
			build_fmftbl(fmscale);
			for (j=0; j<n; ++j)
			{
				rom[i+j*2  ] = (fmftbl[j]&0xFF);
				rom[i+j*2+1] = ((fmftbl[j]>>8)&0xFF);
			}
			return true;
		}
	}
	return false;
}

void build_psgftbl(double scale)
{
	lprintf("psgftbl: ; scale %lf\n", scale);

	int oct = -2, note = 9; // A2
	char *note_name[] = {"C","C#","D","D#","E","F","F#","G","G#","A","A#","B"};
	for (int i=0; i<64; ++i)
	{
		psgftbl[i]=floor(3579545.0/(2*16*ftbl[note]*scale*pow(2.0,oct))+0.5);
		lprintf(" dw %04X ; %s%d\n",(int)psgftbl[i],note_name[note],oct+4);
		++note;
		if (note == 12)
		{
			note = 0;
			++oct;
		}
	}
}

double psgscale = 1.5; //3136.0/2093.0;

bool fix_psgftbl()
{
	// build fmftbl like original
	build_psgftbl(1);
	
	// find psgftbl
	for (int i=0; i+sizeof(psgftbl)-1<rom.size(); ++i)
	{
		int j, n=sizeof(psgftbl)/sizeof(psgftbl[0]);
		for (j=0; j<n; ++j)
			if (abs(rom[i+j*2] - (psgftbl[j]&0xFF)) > 1
			 || rom[i+j*2+1] != ((psgftbl[j]>>8)&0xFF))
				break;
		if (j == n) // found
		{
			build_psgftbl(psgscale);
			for (j=0; j<n; ++j)
			{
				rom[i+j*2  ] = (psgftbl[j]&0xFF);
				rom[i+j*2+1] = ((psgftbl[j]>>8)&0xFF);
			}
			return true;
		}
	}
	return false;
}

int main(int argc, char **args)
{
	if (argc != 3 && argc != 4 && argc != 5 && argc != 6)
		usage();
	
	if (argc > 4)
		sscanf(args[4],"%lf",&fmscale);

	if (argc > 5)
		sscanf(args[5],"%lf",&psgscale);
	
	use_log = (argc>3);

	FILE *f = fopen(args[1],"rb");
	if (!f)
	{
		printf("Can't open rom\n");
		usage();
	}
	fseek(f,0,SEEK_END);
	int flen = ftell(f);
	rom.resize(flen);
	fseek(f,0,SEEK_SET);
	fread(&rom[0],1,flen,f);
	fclose(f);
	
	bool was=false;
	if (fix_fmftbl())
	{
		was = true;
		printf("FM Frequency table successfully fixed!\n");
	}
	if (fix_psgftbl())
	{
		was = true;
		printf("PSG Frequency table successfully fixed!\n");
	}
	if (!was)
	{
		printf("Failed.\n");
		return 0;
	}

	f = fopen(args[2],"wb");
	if (!f)
	{
		printf("Can't open fixed_rom\n");
		usage();
	}
	fwrite(&rom[0],1,flen,f);
	fclose(f);
	printf("Done.");
	return 0;
}